# tests
# Tests for the fastlife module
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 09:12:08 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: __init__.py [] benjamin@bengfort.com $

"""
Tests for the fastlife module
"""

##########################################################################
## Test Constants
##########################################################################

EXPECTED_VERSION = "0.2"


##########################################################################
## Initialization Tests
##########################################################################


class TestInitialization(object):
    def test_sanity(self):
        """
        Test that tests work by confirming 7**3 = 343
        """
        assert 7 ** 3 == 343, "The world went wrong!!"

    def test_import(self):
        """
        Assert that the yellowbrick package can be imported.
        """
        try:
            import fastlife
        except ImportError:
            self.fail("Could not import the fastlife library!")

    def test_version(self):
        """
        Assert that the test version matches the library version.
        """
        try:
            import fastlife as fl

            assert fl.__version__ == EXPECTED_VERSION
        except ImportError:
            self.fail("Could not import the fastlife library!")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# conf
# Fastlife documentation build config file, created by sphinx-quickstart
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 09:28:03 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: conf.py [] benjamin@bengfort.com $

"""
Fastlife documentation build config file, created by sphinx-quickstart

Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

##########################################################################
## Path setup
##########################################################################

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
import os
import sys
sys.path.insert(0, os.path.abspath('..'))

import fastlife

##########################################################################
## Project Information
##########################################################################

project = 'Fastlife'
copyright = '2020, Benjamin Bengfort'
author = 'Benjamin Bengfort'

# The short X.Y version.
version = fastlife.get_version(short=True)
# The full version, including alpha/beta/rc tags.
release = "v" + fastlife.get_version(short=False)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.todo",
    "numpydoc",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# fastlife.sequential
# Implements a Game of Life cellular automata in a single, sequential process.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Mon Jul 27 14:02:53 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: sequential.py [] benjamin@bengfort.com $

"""
Implements a Game of Life cellular automata in a single, sequential process.
"""

##########################################################################
## Imports
##########################################################################

import gzip
import numpy as np
import matplotlib.pyplot as plt

from tqdm import tqdm
from matplotlib.animation import FuncAnimation

from .grid import Grid
from .exceptions import FastlifeError


##########################################################################
## Sequential Life Simulation
##########################################################################


class SequentialLife(object):

    def __init__(self, width=512, height=512):
        # Game of Life has two frames, one for the current timestep and one for the next
        # TODO: load initial state
        self.initialized = False
        self.frames = [Grid(width, height), Grid(width, height)]
        self.frame = 0
        self.now = 0

    @property
    def cframe(self):
        return self.frames[self.frame]

    @property
    def nframe(self):
        return self.frames[1-self.frame]

    def load(self, path):
        """
        Load the simulation from a file on disk.
        """
        grid = self.frames[self.frame]
        with gzip.open(path, 'rb') as f:
            for line in f:
                i, j = tuple(map(int, line.split()))
                grid[i,j] = 1

        self.initialized = True

    def randomize(self, seed=None):
        """
        Create a random initial state from a seed value.
        """
        np.random.seed(seed)
        grid = self.cframe
        grid._world = np.random.randint(2, size=grid.shape)
        self.initialized = True

    def step(self):
        """
        Execute the next step in the simulation and swap the current grid.
        """
        cframe = self.cframe
        nframe = self.nframe

        im, jm = cframe.shape
        for i in range(im):
            for j in range(jm):
                # This is the fastest way to accumulate the number of neighbors.
                ngbrs = cframe.neighborhood_sum(i,j)

                if cframe[i,j] == 1:
                    # Living cell - does it survive or die?
                    if ngbrs < 2 or ngbrs > 3:
                        nframe[i,j] = 0
                    else:
                        nframe[i,j] = 1
                else:
                    # Unoccupied cell - does it get born?
                    if ngbrs == 3:
                        nframe[i,j] = 1
                    else:
                        nframe[i,j] = 0


        # Swap the current frame to the next frame and increment the number of steps
        self.now += 1
        self.frame = 0 if self.frame == 1 else 1

    def run(self, steps=100, progress=True):
        """
        Run the simulation for the specified number of steps from the current state.
        """
        if not self.initialized:
            raise FastlifeError("the game of life simulation has not been initialized")

        if progress:
            for _ in tqdm(range(steps)):
                self.step()
        else:
            for _ in range(steps):
                self.step()

    def animate(self, steps=100):
        """
        Animate the simulation for the specified number of steps.
        """
        if not self.initialized:
            raise FastlifeError("the game of life simulation has not been initialized")

        fig, ax = plt.subplots(figsize=(8,8))
        def plot_frame(idx):
            self.step()
            self.cframe.plot(ax)
            ax.set_title(f"Timestep {self.now}")

        _ = FuncAnimation(
            fig, plot_frame, frames=np.arange(steps),
            repeat=False, init_func=lambda: self.cframe.plot(ax),
        )
        plt.show()



# fastlife
# Experimenting with C code and Python bindings using Conway's Game of Life.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 08:44:49 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: __init__.py [] benjamin@bengfort.com $

"""
Experimenting with C code and Python bindings using Conway's Game of Life.
"""

##########################################################################
## Imports
##########################################################################

from .version import get_version


__version__ = get_version()



# fastlife.utils
# Utility functions and helpers for the fastlife module
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Mon Jul 27 16:13:03 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: utils.py [] benjamin@bengfort.com $

"""
Utility functions and helpers for the fastlife module
"""

##########################################################################
## Imports
##########################################################################

import cProfile

from pstats import Stats
from functools import wraps


try:
    import pandas as pd
except ImportError:
    pd = None


def sprofile(func):
    """
    Decorator that performs a speed/stack profile of the time spent in each function
    call from the stack below the wrapped function. Prints the results when done.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        pr = cProfile.Profile()
        pr.enable()

        result = func(*args, **kwargs)

        pr.disable()
        Stats(pr).strip_dirs().sort_stats('cumulative').print_stats(20)
        return result

    return wrapper


def load_mprofile(path, name=None):
    """
    Load a memory profile from disk for plotting and comparison.
    """
    if pd is None:
        raise ImportError("pandas is required to load the memory profile")

    ref = None
    times, values = [], []
    with open(path, 'r') as f:
        for line in f:
            if line.startswith("CMDLINE"):
                if name is None:
                    name = line.rstrip("CMDLINE").strip()

            if line.startswith("MEM"):
                parts = line.split()
                val, ts = float(parts[1]), float(parts[2])
                if ref is None:
                    ref = ts

                times.append(ts-ref)
                values.append(val)

    return pd.Series(values, index=times, name=name)



# tests.test_grid
# Tests for the grid world helper functionality.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 12:36:38 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: test_grid.py [] benjamin@bengfort.com $

"""
Tests for the grid world helper functionality.
"""

##########################################################################
## Imports
##########################################################################

import pytest
import numpy as np

from fastlife.grid import *
from fastlife.exceptions import *


class TestGrid(object):

    @pytest.mark.parametrize("adjacency", [VON_NEUMANN, MOORE])
    def test_valid_adjacency(self, adjacency):
        """
        Test case-insensitive adjacency parameter
        """
        cases = [
            adjacency, adjacency.title(), adjacency.upper(), adjacency.lower()
        ]
        for case in cases:
            try:
                Grid(adjacency=adjacency)
            except FastlifeValueError:
                pytest.fail(f"'{case}' adjacency is valid but not accepted")

    def test_invalid_adjacency(self):
        """
        Test invalid adjacency raises exception
        """
        with pytest.raises(FastlifeValueError):
            Grid(adjacency="foo")

    def test_indexing(self):
        """
        Test indexing of a grid world
        """
        grid = Grid()
        assert grid._world.sum() == 0
        assert grid[42,19] == 0

        grid[42, 19] = 1
        assert grid[42,19] == 1
        assert grid._world.sum() == 1

    def test_bad_indexing(self):
        """
        Test bad indexing
        """
        grid = Grid()
        with pytest.raises(FastlifeTypeError):
            grid["foo"]

        with pytest.raises(FastlifeTypeError):
            grid["foo"] = 32

        with pytest.raises(FastlifeValueError):
            grid[32, 78] = "foo"

        with pytest.raises(FastlifeValueError):
            grid[32, 78] = 1292

    @pytest.mark.parametrize("adjacency,expected", [
        (VON_NEUMANN, np.asarray([2, 6, 8, 4])),
        (MOORE, np.asarray([2, 3, 6, 9, 8, 7, 4, 1])),
    ])
    def test_neighborhood_internal(self, adjacency, expected):
        """
        Test neighborhood generation inside the world
        """
        grid = Grid(adjacency=adjacency)
        grid[24, 24] = 1
        grid[24, 25] = 2
        grid[24, 26] = 3
        grid[25, 24] = 4
        grid[25, 25] = 5
        grid[25, 26] = 6
        grid[26, 24] = 7
        grid[26, 25] = 8
        grid[26, 26] = 9

        assert (list(grid.neighborhood(25,25)) == expected).all()
        assert (grid.neighborhood_array(25,25) == expected).all()

    @pytest.mark.parametrize("adjacency,expected", [
        (VON_NEUMANN, np.asarray([0, 6, 8, 0])),
        (MOORE, np.asarray([0, 0, 6, 9, 8, 0, 0, 0])),
    ])
    def test_neighborhood_top_corner(self, adjacency, expected):
        """
        Test neighborhood generation on a corner
        """
        grid = Grid(adjacency=adjacency)
        grid[0, 0] = 5
        grid[0, 1] = 6
        grid[1, 0] = 8
        grid[1, 1] = 9

        assert (list(grid.neighborhood(0,0)) == expected).all()
        assert (grid.neighborhood_array(0,0) == expected).all()

    @pytest.mark.parametrize("adjacency,expected", [
        (VON_NEUMANN, np.asarray([2, 6, 0, 4])),
        (MOORE, np.asarray([2, 3, 6, 0, 0, 0, 4, 1])),
    ])
    def test_neighborhood_bottom_row(self, adjacency, expected):
        """
        Test neighborhood generation on a corner
        """
        grid = Grid(adjacency=adjacency)
        grid[98, 49] = 1
        grid[98, 50] = 2
        grid[98, 51] = 3
        grid[99, 49] = 4
        grid[99, 50] = 5
        grid[99, 51] = 6

        with pytest.raises(IndexError):
            grid[100,50] = 8

        assert (list(grid.neighborhood(99, 50)) == expected).all()
        assert (grid.neighborhood_array(99, 50) == expected).all()



# fastlife
# CLI program for running fastlife benchmarks and simulations.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 11:30:13 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: __main__.py [] benjamin@bengfort.com $

"""
CLI program for running fastlife benchmarks and simulations.
"""

##########################################################################
## Imports
##########################################################################

import argparse

from .utils import sprofile
from .version import get_version
from .exceptions import ConsoleError
from .sequential import SequentialLife


##########################################################################
## Module Variables
##########################################################################

PROG = "fastlife"
DESCRIPTION = "experiments in Conway's Game of Life performance"
EPILOG = "please report any bugs on GitHub issues"
VERSION = f"{PROG} v{get_version()}"


##########################################################################
## Console Commands
##########################################################################

def run(args):
    """
    Run a game of life simulation.
    """
    sim = SequentialLife(args.width, args.height)
    if args.file:
        sim.load(args.file)
    else:
        sim.randomize(args.seed)

    runner = sim.animate if args.animate else sim.run
    runner = sprofile(runner) if args.profile else runner
    runner(steps=args.steps)


def bench(args):
    """
    Run game of life benchmarks.
    """
    raise ConsoleError("not implemented yet")



##########################################################################
## Main Method and Argument Parsing
##########################################################################

def make_wide(formatter, width=120, max_help_position=42):
    """
    Increase space between arguments and help text, if possible.
    See: https://stackoverflow.com/questions/5462873/
    """
    try:
        # ensure formatter can be used
        kwargs = {"width": width, "max_help_position": max_help_position}
        formatter(None, **kwargs)

        # return function to create formatter
        return lambda prog: formatter(prog, **kwargs)
    except TypeError:
        # return originally sized formatter
        return formatter


def main():
    cmds = {
        "run": {
            "func": run,
            "description": "run a game of life simulation",
            "args": {
                ("-W", "--width"): {
                    "type": int, "default": 75, "metavar": "W",
                    "help": "the number of columns in the simulation",
                },
                ("-H", "--height"): {
                    "type": int, "default": 75, "metavar": "H",
                    "help": "the number of rows in the simulation",
                },
                ("-f", "--file"): {
                    "type": str, "default": None, "metavar": "PATH",
                    "help": "initialize the simulation from a data file",
                },
                ("-S", "--seed"): {
                    "type": int, "default": None, "metavar": "N",
                    "help": "random seed to load a randomized world with",
                },
                ("-a", "--animate"): {
                    "action": "store_true",
                    "help": "animate the progress of the simulation",
                },
                ("-P", "--profile"): {
                    "action": "store_true",
                    "help": "profile stack calls for the simulation",
                },
                ("-s", "--steps"): {
                    "type": int, "default": 150, "metavar": "T",
                    "help": "maximum number of steps to simulate",
                },
            },
        },
        "bench": {
            "func": bench,
            "description": "run game of life benchmarks",
            "args": {},
        }
    }

    parser = argparse.ArgumentParser(prog=PROG, description=DESCRIPTION, epilog=EPILOG)
    parser.add_argument("-v", "--version", action="version", version=VERSION)
    subparsers = parser.add_subparsers(title="commands")

    for cmd, cmdargs in cmds.items():
        subp = subparsers.add_parser(
            name=cmd,
            description=cmdargs["description"],
            formatter_class=make_wide(argparse.ArgumentDefaultsHelpFormatter),
        )
        subp.set_defaults(func=cmdargs["func"])

        for pargs, kwargs in cmdargs["args"].items():
            if isinstance(pargs, str):
                pargs = (pargs,)
            subp.add_argument(*pargs, **kwargs)

    args = parser.parse_args()
    if "func" in args:
        try:
            args.func(args)
        except ConsoleError as e:
            parser.error(e)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()


# fastlife.grid
# A wrapper for the matrix that describes the game of life grid world.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 12:12:30 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: grid.py [] benjamin@bengfort.com $

"""
A wrapper for the matrix that describes the game of life grid world.
"""

##########################################################################
## Imports
##########################################################################

import numpy as np
import matplotlib.pyplot as plt

from .exceptions import FastlifeValueError, FastlifeTypeError


VON_NEUMANN = "von neumann"
MOORE = "moore"

VNIP = np.asarray([-1, 0, 1, 0])
VNJP = np.asarray([0, 1, 0, -1])
MRIP = np.asarray([-1, -1, 0, 1, 1, 1, 0, -1])
MRJP = np.asarray([0, 1, 1, 1, 0, -1, -1, -1])


class Grid(object):
    """
    The game of life world is a 2-dimensional grid that is represented by a numpy matrix
    whose cells are False (dead) or True (alive). The grid provides wrapper functions
    for neighborhood calculations, partitioning, and other operations required for
    parallel implementations of cellular automata.

    Parameters
    ----------
    width, height : int
        The shape of the underlying grid world.

    adjacency : str, default: "moore"
        The type of neighborhood, either "von neumann" or "moore"
    """

    __slots__ = ["_world", "_adjacency"]

    def __init__(self, width=100, height=100, adjacency=MOORE):
        self._world = np.zeros((height, width), dtype=np.int8)
        self.adjacency = adjacency

    @property
    def adjacency(self):
        return self._adjacency

    @adjacency.setter
    def adjacency(self, val):
        val = val.lower().replace("_", " ").strip()
        if val != MOORE and val != VON_NEUMANN:
            raise FastlifeValueError(f"'{val}' is not a valid adjacency")
        self._adjacency = val

    @property
    def shape(self):
        return self._world.shape

    def neighborhood(self, i, j):
        """
        Returns the neighborhood of the cell specified at the x, y coords as a generator
        of values that the caller can iterate over. This mechanism is the fastest way
        to collect all of the individual values of the neighbors. Note that neighbors
        are returned from the top neighbor (e.g. the neighbor directly above) clockwise.

        Parameters
        ----------
        i, j : int
            The position of the cell in the world to get the neighbors for.
        """
        ip, jp = (MRIP, MRJP) if self.adjacency == MOORE else (VNIP, VNJP)

        for id, jd in zip(ip, jp):
            try:
                yield self._world[i+id,j+jd]
            except IndexError:
                yield 0

    def neighborhood_sum(self, i, j):
        """
        Returns the sum of all cells in the neighborhood of i,j without yielding them
        as a generator. This is the fastest performing method to get the total number of
        neighbors if the cellular automata uses 1s and 0s for state.

        Parameters
        ----------
        i, j : int
            The position of the cell in the world to get the neighbors for.
        """
        total = 0
        ip, jp = (MRIP, MRJP) if self.adjacency == MOORE else (VNIP, VNJP)

        for id, jd in zip(ip, jp):
            try:
                total += self._world[i+id,j+jd]
            except IndexError:
                continue
        return total

    def neighborhood_array(self, i, j):
        """
        Returns the neighborhood of the cell specified at the x, y coords as a numpy
        array whose shape is determined by the adjacency. Note that this is not as fast
        as using the generator or sum methods above and should only be used if more
        extensive computations are required for the neighborhood array. Neighbors are
        returned from the top neighbor (e.g. the neighbor directly above) clockwise.

        Parameters
        ----------
        i, j : int
            The position of the cell in the world to get the neighbors for.
        """
        if self.adjacency == MOORE:
            ip, jp = MRIP, MRJP
            vals= np.zeros(8)
        else:
            ip, jp = VNIP, VNJP
            vals = np.zeros(4)

        im, jm = self._world.shape
        for v, (id, jd) in enumerate(zip(ip, jp)):
            ic, jc = i+id, j+jd
            if ic >= 0 and jc >=0 and ic < im and jc < jm:
                vals[v] = self._world[ic,jc]

        return vals

    def plot(self, ax=None):
        if ax is None:
            _, ax = plt.subplots(figsize=(8,8))
        ax.imshow(self._world, aspect="equal")
        ax.set_xticks([])
        ax.set_yticks([])
        return ax

    def __getitem__(self, ij):
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise FastlifeTypeError("specify i, j position as a two-tuple")
        return self._world[ij[0], ij[1]]

    def __setitem__(self, ij, val):
        if not isinstance(ij, tuple) or len(ij) != 2:
            raise FastlifeTypeError("specify i, j position as a two-tuple")

        if not isinstance(val, (int, np.int8)) or val > 127 or val < -128:
            raise FastlifeValueError("invalid game of life value")
        self._world[ij[0], ij[1]] = val



# fastlife.exceptions
# Exception and warnings hierarchy for use with fastlife.
#
# Author:   Benjamin Bengfort <benjamin@bengfort.com>
# Created:  Sun Jul 26 11:38:51 2020 -0400
#
# Copyright (C) 2020 Bengfort.com
# For license information, see LICENSE
#
# ID: exceptions.py [] benjamin@bengfort.com $

"""
Exception and warnings hierarchy for use with fastlife.
"""

##########################################################################
## Exception Hierarchy
##########################################################################

class FastlifeError(Exception):
    pass


class FastlifeTypeError(FastlifeError, TypeError):
    pass


class FastlifeValueError(FastlifeError, ValueError):
    pass


class ConsoleError(FastlifeError):
    pass



##########################################################################
## Warnings Hierarchy
##########################################################################

class FastlifeWarning(UserWarning):
    pass